#include <iostream>
#include <fstream>
#include <set>
#include <functional>
#include <vector>
#include "main.h"
#include "data.h"
#include "misc.h"

// ------- class vertices ----------------------------

int vertices::n=0;  
int vertices::d=0;  
vertices_lookup vertices::lookup=vertices_lookup();

void vertices::set_dimensions(int N, int D) 
{ 
  n=N; d=D; 
  lookup.generate_tables(n,d);
}


const simplex vertices::orden() const
{
  simplex s=1;
  vertex k=1,l;
  const_iterator l_iterator=begin();  
  for (vertex i=1; i<=d; ++i) {    
    l=*l_iterator;    
    for (vertex j=k; j<=l-1; ++j)
      s=s+lookup.get_binomial(n-j,d-i);    
      //      s=s+binomial(n-j,d-i);    
    k=l+1;    
    ++l_iterator;  
  }  
  return(s);
}

const vertices & vertices::reverse_orden(const simplex & S) 
{
  (*this) = lookup.reverse_orden(S);
  return(*this);  
}

std::ostream & operator << (std::ostream & out, const vertices & v)
{
  out << *( v.begin() );  
  vertices::const_iterator i=v.begin();   i++;  
  for (; i!=v.end(); ++i)
    out << "_" << *i;
  return(out);
}


bool vertices_order::operator()(const vertices & a, const vertices & b) const
{
  size_t n1=a.size(), n2=b.size();
  if (n1<n2) return(true);
  if (n1>n2) return(false);
  // a,b have same size
  for (vertices::iterator i=a.begin(), j=b.begin();
       i!=a.end() && j!=b.end(); ++i, ++j) {    
    if (*i < *j) return(true);  
    if (*i > *j) return(false);
  }
  // a == b
  return(false);
}


bool operator==(const vertices & a, const vertices & b) 
{
  return( std::set<vertex,std::less<vertex> >(a) == 
          std::set<vertex,std::less<vertex> >(b) );
}


// ------- class vertices_lookup --------------------

const vertices & vertices_lookup::reverse_orden(const simplex & S) const 
{
  return( ReverseOrden[S-1] );
}

// this one is messy, but it only reverses vertices::orden()
vertices vertices_lookup::manual_reverse_orden(const simplex & S) const
{
  vertices result;  
  simplex s=S;
  simplex b;  
  vertex i,j,l=0,k;
  for (k=1; k<d; k++) {    
    l++;  i=l;   j=1; 
    b=binomial(n-l,d-k);
    while (s>b && b>0) { 
      j++;  l++;
      s=s-b;  
      b=binomial(n-l,d-k);
    }
    result.insert(result.begin(),l);
  }
  result.insert(result.begin(),s+l);
  return(result);
} 

// tweaked for special (e.g. negative) values 
int vertices_lookup::get_binomial(int i, int j) const 
{ 
  if (i>=0 && i<=n && j>=0 && j<=d && j<=i)
    return(fast_binomial[i][j]); 
  else 
    return(1);
}  


void vertices_lookup::generate_tables(int N, int D)
{
  n=N;  d=D;  
  fast_binomial.erase(fast_binomial.begin(),fast_binomial.end());
  for(int i=0; i<=n; ++i) {
    std::vector<int> row;    
    for(int j=0; j<=i && j<=d; j++)
      row.push_back( binomial(i,j) );
    fast_binomial.push_back( row );
  }
  ReverseOrden.erase(ReverseOrden.begin(),ReverseOrden.end());  
  for(int s=1; s<=binomial(n,d); ++s)
    ReverseOrden.push_back(manual_reverse_orden(s));
}




// ------- class compact_simplices -------------------

compact_simplices::compact_simplices()  
{ 
  reserve(50);  
}
 
compact_simplices::~compact_simplices()
{
}
 
std::istream & operator >> (std::istream & in, compact_simplices & t)
{
  int bracketlevel=0;
  simplex s;
  char ch;
  do {
    in >> ch;
    switch (ch) {
    case '0'...'9': 
      in.putback(ch);  // oops :-)
      in >> s;
      t.push_back(s);
      break;
    case '[': 
      bracketlevel++;
      break;
    case ']': 
      bracketlevel--;
      break;
    }
  } while (ch!=';' && bracketlevel>=0 && !in.eof());
  return(in);  
}

std::ostream & operator << (std::ostream & out, const compact_simplices & t)
{
  out << "[" << *( t.begin() );  
  for (compact_simplices::const_iterator i=t.begin()+1; i!=t.end(); ++i)
    out << "," << *i;
  out << "]";
  return(out);
}

const bool operator==(const compact_simplices & a, const compact_simplices & b)
{
  // friendship is not inherited
  return( std::vector<simplex>(a) == std::vector<simplex>(b)  );
}

hash_value compact_simplices::hash_function() const
{
  unsigned long int result=0; // ulong to avoid overflow
  int n=1;  
  for (const_iterator i=begin(); i!=end(); ++i, n+=151)
    result = (result + n* (*i)) % hash_max;
  return(result);
}


bool compact_simplices::starshaped() const
{
  bool result = true;  
  const simplex origin = simplex(1);  
  vertices v;  
  for (const_iterator i=begin(); i!=end(); i++) {
    v.reverse_orden(*i);
    result=result && (find(v.begin(),v.end(),origin)!=v.end());
  }
  return(result);           
}




// ------- class simplices ---------------------------

simplices::simplices(const compact_simplices & s)
  :compact_simplices(s)
{
  // de-compress vertex information
  reverse_orden();  
}

simplices::simplices(const std::set<vertices,vertices_order> & s)
{
  v.erase(v.begin(), v.end());
  copy(s.begin(), s.end(), back_inserter(v) );
  orden();  
}


simplices::~simplices()
{
}

void simplices::orden()
{
  erase(begin(),end());  
  for (std::vector<vertices>::const_iterator i=v.begin(); i!=v.end(); i++)
    push_back( (*i).orden() );  
  sort(begin(),end());  
}


void simplices::reverse_orden()
{
  v.erase(v.begin(),v.end());  
  for (const_iterator i=begin(); i!=end(); i++)
    v.push_back( vertices().reverse_orden(*i) );
}

std::ostream & operator << (std::ostream & out, const simplices & s)
{
  out << "[" << *( s.v.begin() );  
  std::vector<vertices>::const_iterator i=s.v.begin();
  i++;  
  for (; i!=s.v.end(); ++i)
    out << " , " << *i;
  out << "]";
  return(out);
}



// ------- class flip --------------------------------

flip::flip()
{
  deltaplus.reserve(10);
  deltaminus.reserve(10);
}
  
flip::flip(std::istream & in)
{
  deltaplus.reserve(10);
  deltaminus.reserve(10);
  char ch;
  std::vector<vertices> * store_to = NULL;  
  for (int i=1; i<=2; i++) {    
    switch(i) {      
      case 1: store_to = &deltaplus;  break;
      case 2: store_to = &deltaminus; break; 
    }
    vertex v;
    vertices s;  
    int bracketlevel=1;  
    do {
      in >> ch;
      switch (ch) {
      case '0'...'9': 
	in.putback(ch);  // oops :-)
	in >> v;      
	s.insert(s.begin(),v);
	break;
      case '{':
      case '[': 
	if (bracketlevel==1) s.erase(s.begin(),s.end());      
	bracketlevel++;
	break;
      case '}':
      case ']': 
	if (bracketlevel==2) store_to->push_back(s);      
	bracketlevel--;      
	break;
      }
    } while (ch!=';' && bracketlevel>=1 && !in.eof());
    while (ch!='{' && ch!=';' && ch!=']' && !in.eof())
      in >> ch;
  }  
}

flip::flip(const flip & copyfrom)
{
  deltaplus.reserve(10);
  deltaminus.reserve(10);
  (*this)=copyfrom;  
}

flip::~flip()
{
}

flip & flip::operator =(const flip & copyfrom)
{
  deltaplus  = copyfrom.get_deltaplus();
  deltaminus = copyfrom.get_deltaminus();  
  return(*this);  
}

std::ostream & operator << (std::ostream & out, const flip & f)
{
  out << "< ";  
  for (std::vector<vertices>::const_iterator i=f.get_deltaplus().begin(); 
       i!=f.get_deltaplus().end(); ++i) {      
    std::cout << *i << " ";
  }      
  out << "| ";  
  for (std::vector<vertices>::const_iterator i=f.get_deltaminus().begin(); 
       i!=f.get_deltaminus().end(); ++i) { 
    std::cout << *i << " ";    
  }      
  out << ">\n";
  return(out);
}

void flip::mirror()
{
  swap(deltaplus,deltaminus);  
}




// ------- class flips -------------------------------

flips::flips()
{
}

flips::~flips()
{
}

std::istream & operator>>(std::istream & in, flips & f)
{
  char ch;
  int n,d;  
  // read n := <blah>;
  do {
    in >> ch;
    switch (ch) {
    case '0'...'9': 
      in.putback(ch);  // oops :-)
      in >> n;
      break;
    }
  } while (ch!=';' && !in.eof());
  // read d := <blah>;
  do {
    in >> ch;
    switch (ch) {
    case '0'...'9': 
      in.putback(ch);  // oops :-)
      in >> d;
      break;
    }
  } while (ch!=';' && !in.eof());
  vertices().set_dimensions(n,d);  
  // read flip:= [[{{<flip1>},[<flip2>,...]];
  do {
    in >> ch;
    switch (ch) {
    case '{': 
      flip new_flip(in);	
      f.push_back(new_flip);
      new_flip.mirror();
      f.push_back(new_flip);
      break;
    }
  } while (ch!=';' && !in.eof()); 
  return(in);  
}


 
// ------- class goodcircuit -------------------------


goodcircuit::goodcircuit(const simplices & s, const flip & f)
  :supporter(f)
{    
  const std::vector<vertices> & deltaplus = f.get_deltaplus();  
  const std::vector<vertices> & v = s.get_vertices();  
  supported.reserve(deltaplus.size());
  good=true;  
  

  // find simplices that include members of deltaplus 
  std::vector<vertices> empty_result;  
  for (size_t n=0; n<deltaplus.size(); n++) {
    supported.push_back(empty_result);  supported[n].reserve(10);  
    bool found_match=false;    
    for (std::vector<vertices>::const_iterator i=v.begin(); i!=v.end(); ++i) {
      if (includes( (*i).begin(), (*i).end(), 
		    deltaplus[n].begin(), deltaplus[n].end() )) {
	found_match=true;
	supported[n].push_back(*i);	
      }
    }
    good=good && found_match;
    if (!good) return;    
  }

  // check that all links are equal
  for (size_t n=0; n<deltaplus.size(); n++) {
    std::set<vertices,vertices_order> l;  
    for (std::vector<vertices>::const_iterator i=supported[n].begin();
    	 i!=supported[n].end(); ++i) {      
      vertices one_simplex;    
      std::insert_iterator<vertices> 
	ins(one_simplex,one_simplex.begin());    
      set_difference( (*i).begin(), (*i).end(), 
 		      deltaplus[n].begin(), deltaplus[n].end(),
 		      ins );
      l.insert(l.begin(),one_simplex);      
    }    
    link.push_back(l);    
    if (n>0 && link[0]!=link[n]) {
      good=false;
      return;
    }   				    
  }
}


void goodcircuit::do_flip(const simplices & s, const flip & f)
{
  bistellarneighbor.erase(bistellarneighbor.begin(), bistellarneighbor.end());
  
  const std::set<vertices,vertices_order> & use_link = link[0];
  const std::vector<vertices> & v = s.get_vertices();  
  const std::vector<vertices> & deltaplus  = f.get_deltaplus();  
  const std::vector<vertices> & deltaminus = f.get_deltaminus();  
  
  // start with all simplices 
  std::set<vertices,vertices_order> tri;  
  copy(v.begin(), v.end(), inserter(tri,tri.begin()) );
  
  // remove all suported simplices
  std::set<vertices,vertices_order> to_remove;
  
  for (std::set<vertices,vertices_order>::const_iterator i=use_link.begin();
       i!=use_link.end(); ++i) 
    for (std::vector<vertices>::const_iterator j=deltaplus.begin();
	 j!=deltaplus.end(); ++j) {    
      vertices one_simplex;
      set_union( (*j).begin(), (*j).end(),
		 (*i).begin(), (*i).end(),
		 inserter(one_simplex,one_simplex.begin()) );
      to_remove.insert(to_remove.begin(),one_simplex);            
    }
  set_difference( tri.begin(), tri.end(), 
		  to_remove.begin(), to_remove.end(),
		  inserter(bistellarneighbor,bistellarneighbor.begin()),
		  vertices_order() );
    
  // now add the flip of the supported simplices
  for (std::set<vertices,vertices_order>::const_iterator i=use_link.begin();
       i!=use_link.end(); ++i)
    for (std::vector<vertices>::const_iterator j=deltaminus.begin();
	 j!=deltaminus.end(); j++) {
      vertices one_simplex;
      set_union( (*j).begin(), (*j).end(),
		 (*i).begin(), (*i).end(),
		 inserter(one_simplex,one_simplex.begin()) );
      bistellarneighbor.insert( bistellarneighbor.begin(), one_simplex );
    }  
}













