#include <time.h>
#include <iostream>
#include <fstream>
#include <set>
#include <functional>
#include <vector>
#include "main.h"
#include "data.h"

class triangulations: public std::vector<compact_simplices>
{
private:
  compact_simplices no_triangulation_instance;
  compact_simplices::const_iterator no_triangulation;  
  std::vector<size_t> hash_list;  
  mutable size_t freespace;  
  size_t max_simplices;  
protected:
  bool is_new(const compact_simplices &) const;
  void add_triangulation(const compact_simplices &);  
  bool printworthy(const compact_simplices & s) const;  
public:
  mutable int collisions;  
  triangulations();
  void add_neighbours(const simplices &, const flips &);    
  friend std::ostream & operator <<(std::ostream &, const triangulations &);  
  void add_triang_if_new(const compact_simplices &);  
  void save() const;  
};

triangulations::triangulations()
  :hash_list(hash_max,hash_max),
   max_simplices(0),
   collisions(0) 
{}  

// faster decision than find(begin(),end(),t)==end()
bool triangulations::is_new(const compact_simplices & t) const
{
  hash_value n=t.hash_function();
  for (hash_value i=n; i<hash_max; ++i)
    if (hash_list[i]==hash_max) {
      freespace=i;      
      return(true); // empty place in hash
    }  
    else if (*(begin()+hash_list[i])==t) 
      return(false); // found ourselves in hash
    else collisions++;  
  // wrap to beginning of hash
  for (size_t i=0; i<n; ++i)
    if (hash_list[i]==hash_max) {
      freespace=i;      
      return(true); // empty place in hash
    }  
    else if (*(begin()+hash_list[i])==t) 
      return(false); // found ourselves in hash
    else collisions++;  
  // all places in has are filled with other triangulations
  std::cerr << "Hash list full" << std::endl;
  exit(0);
}

// you are only allowed to add after checking is_new()
void triangulations::add_triangulation(const compact_simplices & new_triang)
{
  push_back(new_triang);   
  hash_list[freespace] = size()-1;  
  if (new_triang.size()>max_simplices && new_triang.starshaped()) 
    max_simplices=new_triang.size();  
}

void triangulations::add_triang_if_new(const compact_simplices & new_triang)
{
  if (is_new(new_triang)) {	
    add_triangulation(new_triang);   
    //	std::cout << new_triang << "\n";   
  }      
}

void triangulations::add_neighbours(const simplices & s, const flips & all_f)
{
  for (flips::const_iterator f=all_f.begin(); f!=all_f.end(); ++f) {
    goodcircuit goody(s,*f);    
    if (goody.is_good()) {
      goody.do_flip(s,*f);
      compact_simplices new_triang=goody.get_neighbor();   
      add_triang_if_new(new_triang);      
    }    
  } 
}

void triangulations::save() const
{
  // write data
  std::ofstream outfile("puntos.output");
  if (!(outfile)) {
    std::cerr << "Cant write results\n";
    exit(1);
  }    
  outfile << *this;  
  outfile.close();  
}


bool triangulations::printworthy(const compact_simplices & s) const
{
  bool good=true;
  // good = good && (s.size()==max_simplices);
  good = good && s.starshaped();  
  return(good);  
}


std::ostream & operator<<(std::ostream & out, const triangulations & triang)
{
  // output all found triangulations
  //   out << "triangulations:=[" << *( triang.begin() );  
  //   for (triangulations::const_iterator i=triang.begin(); 
  //        i!=triang.end(); ++i)
  //     out << ",\n" << *i;
  //   out << "];\n";  

  // output all starshaped triangulations
  triangulations::const_iterator i=triang.begin();
  while (!triang.printworthy(*i) && i!=triang.end()) ++i; 
  if (i==triang.end()) return(out); // no output  
  out << "triangulations:=[" << *i;  
  ++i;  
  for (; i!=triang.end(); ++i)
    if (triang.printworthy(*i))    
      out << ",\n" << *i;
  out << "];\n";  

  return(out);
}




class timer
{  
private:
  time_t initial_time, last_output_time;
  time_t alarm;  
public:
  timer() :alarm(5) { reset(); }
  virtual ~timer() {};
  void set_period(time_t dt)  { alarm=dt; }
  bool activate();  
  void reset() { initial_time=time(0); last_output_time=initial_time; }     
  time_t now_passed() { return(time(0)-initial_time); }       
};

bool timer::activate() 
{
  time_t delta_t = time(0)-last_output_time;
  if (delta_t<alarm) 
    return(false);
  // else
  last_output_time+=delta_t;       
  return(true);
}




int main(void)
{
  compact_simplices seed;
  flips all_flips;  

  // read data
  std::ifstream infile("puntos.input");
  if (!(infile)) {
    std::cerr << "File not found\n";
    return(1);
  }    
  infile >> seed;
  infile >> all_flips; 
  infile.close();  

  // print data so we can see that it worked
  std::cout << "\n" << "Seed triangulation: " << seed << "\n";
  {    
    std::cout << "Vertices of seed triangulation: ";    
    simplices s(seed);    
    std::cout << s << "\n";    
  }       
  std::cout << "All flips:\n";  
  for (flips::const_iterator i=all_flips.begin(); i!=all_flips.end(); ++i)
    std::cout << *i;



  // do the actual work
  timer t; t.set_period(300);  
  triangulations triangs;
  triangs.add_triang_if_new(seed); 
  size_t position=0;
  do {   
    //    std::cout << triangs;  
    triangs.add_neighbours(triangs[position],all_flips);
    position++;        
    if (t.activate()) {
      std::cout 
        << "Found " << triangs.size() << " triangulations "
	<< "and checked " << position
	<< " in " << t.now_passed() << "s with " 
	<< triangs.collisions << " collisions in hash" << std::endl;
      triangs.save();      
    }   	
  } while (position!=triangs.size());
  std::cout << "Finished: found " << triangs.size() << " triangulations in " 
       << t.now_passed() << "s\n";
  triangs.save();      
  

  
  return(0);
}

